#include <stdlib.h>


#ifndef PC_TEST
#define LOW_ADDRESS      ((( (unsigned) ( & end ) ) + 0x100 )) 
#define HIGH_ADDRESS     ( 1 << 26 )


#include "xparameters.h"	
#include "xcache_l.h"
#else
#define BLOCK_SIZE 4096
unsigned char block [ BLOCK_SIZE ] ;
unsigned LOW_ADDRESS ; 
unsigned HIGH_ADDRESS ; 
#endif

#define XOR_KEY 0x12345678
#define RAND_TABLE_SIZE		4093
unsigned rand_table [ RAND_TABLE_SIZE ];
unsigned rand_index , seed ;
extern unsigned end ;

typedef void (* Target_Fn) ( int , int ) ;

static unsigned rand32 ( void )
{
    if ( rand_index >= RAND_TABLE_SIZE )
	 {
	    rand_index = 0;
    }
	 return rand_table [ rand_index ++ ] ^ XOR_KEY;
}

static void print_hex ( unsigned val )
{
    unsigned i ;
    unsigned char s [ 2 ] ;

    s [ 1 ] = '\0' ;
    for ( i = 0 ; i < 32 ; i += 4 )
    {
        unsigned nibble = (( val << i ) >> 28 ) & 0xf ;
        
        if ( nibble < 10 )
        {
            s [ 0 ] = nibble + 48 ;
        } else {
            s [ 0 ] = nibble + 97 - 10 ;
        }
        print ( s ) ;
    }
}

static int serial_tstc ( void )
{
    return (( XIo_In32 ( XPAR_UARTLITE_0_BASEADDR + 0x8 ) & 0x01 ) != 0 ) ;
}

static inline void update_status ( unsigned addr )
{
    if (( addr & 0xffff ) != 0 ) return ;
	 print_hex ( addr );
	 print ( "\x08\x08\x08\x08\x08\x08\x08\x08" );
#ifndef PC_TEST
    while ( serial_tstc () )
    {
		  if ( inbyte () == 'R' )
		  {
             Target_Fn       fn ;

             fn = (Target_Fn) 0xfffff000 ;
				 fn ( 0, 0 );
		  }
    }
#endif	 
}


static int back_test ( int read_size , int read_index ) ;

int main ( void )
{
    unsigned        pass , temp = 0 ;

#ifdef PC_TEST
    LOW_ADDRESS = (unsigned) & block [ 0 ] ;
    HIGH_ADDRESS = LOW_ADDRESS + BLOCK_SIZE ;
#endif
    print ( "ppcmemtest\r\n" ) ;
    print ( "testing " ) ;
    print_hex ( LOW_ADDRESS ) ;
    print ( " to " ) ;
    print_hex ( HIGH_ADDRESS ) ;
    print ( "\r\n" ) ;
    if ( LOW_ADDRESS < (unsigned) & end )
    {
        print ( "This overlaps the program!\r\n" ) ;
#ifdef PC_TEST
        print ( "oh well...\r\n" ) ;
#else
        return 1 ;
#endif
    }

#ifndef PC_TEST
    asm volatile (
            /* disable I and D caches */
            "isync\n"
            "mtspr 1008,%0\n" /* HID0 */
            "sync\n"
            "isync\n" 
            "iccci %0,%0\n" : "=&r"(temp) : "0"(temp) ) ;


    XCache_EnableICache(0x80000000); 
#endif

    srand ( 26 );
    for ( rand_index = 0; rand_index < RAND_TABLE_SIZE ; rand_index ++ )
	 {
	     rand_table [ rand_index ] = rand() ^ ( rand() << 16 ) ^ rand_index 
				^ ( rand_index << 10 ) ^ ( rand_index << 20 ) ;
	 }
    rand_index = 0 ;
    pass = 0 ;
    while ( 1 )
    {
        int error = 0;
        unsigned addr;

        print ( "pass " ) ;
        print_hex ( pass ) ;
        print ( " - write " ) ;

		  rand_index = seed = (( pass * 73 ) % RAND_TABLE_SIZE ) ;
        for ( addr = LOW_ADDRESS ; addr < HIGH_ADDRESS ; addr += 4 )
        {
            unsigned * ptr = ((unsigned *) addr) ;

            ptr [ 0 ] = rand32 () ;
            update_status ( addr ) ;
        }
        print ( "read " ) ;
        error += back_test ( 4 , 0 ) ;
        error += back_test ( 2 , 0 ) ;
        error += back_test ( 2 , 1 ) ;
        error += back_test ( 1 , 0 ) ;
        error += back_test ( 1 , 1 ) ;
        error += back_test ( 1 , 2 ) ;
        error += back_test ( 1 , 3 ) ;

        if ( error )
        {
            print ( "\r\nTotal " ) ;
            print_hex ( error ) ;
            print ( " error(s)\r\n\r\n" ) ;
            return 1 ;
        } else {
            print ( "ok" ) ;
        }
        print ( "\r" ) ;
        pass ++ ;
    }
}

static int back_test ( int read_size , int read_index )
{
    int error = 0 ;
    unsigned addr ;

    rand_index = seed ;
    for ( addr = LOW_ADDRESS ; addr < HIGH_ADDRESS ; addr += 4 )
    {
        unsigned * ptr = ((unsigned *) addr) ;
        unsigned expect = rand32 () ;
        volatile unsigned cap ;

        if ( read_size == 1 )
        {
            cap = (unsigned) (( (unsigned char *) ptr ) [ read_index ] ) ;
#ifdef PC_TEST
            expect = expect >> ( read_index * 8 ) ;
            expect &= 0xff ;
#else
            expect = expect << ( read_index * 8 ) ;
            expect = expect >> 24 ;
#endif
        } else if ( read_size == 2 )
        {
            cap = (unsigned) (( (unsigned short *) ptr ) [ read_index ] ) ;
#ifdef PC_TEST
            expect = expect >> ( read_index * 16 ) ;
            expect &= 0xffff ;
            /*if ( read_index == 0 )
            {
                expect = kkk(( expect & 0xff ) << 8 ) | 
                        (( expect & 0xff00 ) >> 8 ) ;
            } else {
                expect = (( expect & 0xff0000U ) >> 8 ) | 
                        (( expect & 0xff000000U ) >> 24 ) ;
            }*/
#else
            expect = expect << ( read_index * 16 ) ;
            expect = expect >> 16 ;
#endif
        } else {
            cap = ptr [ 0 ] ;
        }

        if ( cap != expect )
        {
            if ( error == 0 )
            {
                print ( "\r\nERROR at " ) ;
                print_hex ( addr ) ;
                print ( " expected " ) ;
                print_hex ( expect ) ;
                print ( " got " ) ;
                print_hex ( cap ) ;
                print ( " read size " ) ;
                print_hex ( read_size ) ;
                print ( " index " ) ;
                print_hex ( read_index ) ;
                print ( " " ) ;
            }
            error ++ ;
        }
        update_status ( addr ) ;
    }
    return error ;
}


